#' Main DEgenes Hunter Function
#'
#' This function allows you to perform the DEG analysis with different 
#' algorithms.
#' @param raw Dataframe with raw counts per sample and feature
#' @param target set with control and treatment information
#' @param external_DEA_data external DEA set 
#' @param count_var_quantile quantile variance threshold
#' @param output_files otput files path
#' @param reads set of reads
#' @param minlibraries minimum of libraries to use a set of data
#' @param filter_type filter type to ba applied. Allowed: "separate" or "global"
#' @param p_val_cutoff p-value threshold
#' @param lfc minimum logFold Change
#' @param modules modules to be executed. Allowed: DESeq2 (D), NOISeq (N), 
#' Limma (L), EdgeR (E), WGCNA (W)
#' @param minpack_common minimum of pack that must be significant to tag 
#' a gene as significant
#' @param model_variables custom model
#' @param numerics_as_factors transform numeric values to factors. Default: TRUE
#' @param string_factors string factors for WGCNA
#' @param numeric_factors numeric factors for WGCNA
#' @param WGCNA_memory see WGCNA package
#' @param WGCNA_norm_method see WGCNA package
#' @param WGCNA_deepsplit see WGCNA package
#' @param WGCNA_min_genes_cluster see WGCNA package
#' @param WGCNA_detectcutHeight see WGCNA package
#' @param WGCNA_mergecutHeight see WGCNA package
#' @param WGCNA_all see WGCNA package
#' @param WGCNA_blockwiseNetworkType see WGCNA package
#' @param WGCNA_blockwiseTOMType see WGCNA package
#' @param WGCNA_minCoreKME see WGCNA package
#' @param WGCNA_minCoreKMESize see WGCNA package
#' @param WGCNA_minKMEtoStay see WGCNA package
#' @param multifactorial specify interaction/effect when multifactorial design
#' @return expression analysis result object with studies performed
#' @keywords method
#' @importFrom rlang .data
#' @export
#' @examples
#' data(toc)
#' data(target)
#' degh_out <- main_degenes_Hunter(raw=toc, target=target, modules="D")

main_degenes_Hunter <- function(
    raw = NULL,
    target = NULL,
    count_var_quantile = 0,
    external_DEA_data = NULL,
    output_files = getwd(),
    reads = 2,
    minlibraries = 2,
    filter_type = "separate",
    p_val_cutoff = 0.05,
    lfc = 1,
    modules ="DELNW",
    minpack_common = 4,
    model_variables = "",
    numerics_as_factors = TRUE,
    string_factors = "",
    numeric_factors = "",
    WGCNA_memory = 5000,
    WGCNA_norm_method = "DESeq2",
    WGCNA_deepsplit = 2,
    WGCNA_min_genes_cluster = 20,
    WGCNA_detectcutHeight = 0.995,
    WGCNA_mergecutHeight = 0.25,
    WGCNA_all = FALSE,
    WGCNA_blockwiseNetworkType = "signed",
    WGCNA_blockwiseTOMType = "signed",
    WGCNA_minCoreKME = 0.5,
    WGCNA_minCoreKMESize = NULL,
    WGCNA_minKMEtoStay = 0.2,
    multifactorial = ""
  ){
    modified_input_args <- check_input_main_degenes_Hunter(raw, 
      minlibraries, reads, external_DEA_data, modules, model_variables,
      active_modules, WGCNA_all, minpack_common, target, 
      string_factors, numeric_factors, multifactorial)
    modules <- modified_input_args[['modules']]
    active_modules <- modified_input_args[['active_modules']]
    minpack_common <- modified_input_args[['minpack_common']]
    raw <- modified_input_args[['raw']]
    reads <- modified_input_args[['reads']]
    target <- modified_input_args[['target']]
    final_main_params <- list(
      'minpack_common' = minpack_common,
      'p_val_cutoff' = p_val_cutoff,
      'lfc' = lfc,
      'modules' = modules
    )

    ############################################################
    ##                         I/O DATA                       ##
    ############################################################


    # Infer replicates and group index from target
    index_control_cols <- as.character(target$sample[target$treat == "Ctrl"])
    index_treatmn_cols <- as.character(target$sample[target$treat == "Treat"])
    replicatesC <- length(index_control_cols)
    replicatesT <- length(index_treatmn_cols)
    design_vector <- c(rep("C", replicatesC), rep("T", replicatesT))
    sample_groups <- data.frame(class = design_vector, 
                                name = c(index_control_cols, 
                                         index_treatmn_cols))   
    # Check if there are enough replicates for specified method
    if((replicatesC < 2) | (replicatesT < 2)) 
       stop(paste0('At least two replicates per class (i.e. treatment and',
                   ' control) are required\n'))


    if(!is.null(target) & grepl("W", modules)) {
      target_numeric_factors <- build_design_for_WGCNA(target, 
           numeric_factors=numeric_factors)
      target_string_factors <- build_design_for_WGCNA(target, 
          string_factors=string_factors)
    }

    if(numerics_as_factors == TRUE) { 
      # Now coerce the targets to factors for the multifactorial analysis
      invisible(lapply(seq(ncol(target)),function(i){
        target[,i] <<- as.factor(target[,i])
      }))
      # target <-  data.frame(sapply(target, as.factor), 
        # stringsAsFactors=TRUE)

    }
    model_formula_text <- prepare_model_text(model_variables, 
                                             multifactorial)
   
    if(multifactorial != ""){
      target <- prepare_target_for_multifactorial(target, multifactorial)
    }
    # Prepare count table for analysis
    #Indexing selected columns from input count dataframe   
    raw <- raw[c(index_control_cols,index_treatmn_cols)]
    raw[is.na(raw)] <- 0 # Substitute NA values

    raw_filter <- filter_count(reads, minlibraries, raw, filter_type, 
                               index_control_cols, index_treatmn_cols)

    var_filter <- filter_by_variance(raw_filter, 
                                     q_filter = count_var_quantile, 
                                     target = target)
    raw_filter <- var_filter[["fil_count_mtrx"]]

    
    ############################################################
    ##             PERFORM EXPRESION ANALYSIS                 ##
    ############################################################
    dir.create(output_files)
   
    exp_results <- perform_expression_analysis(modules, replicatesC, 
                     replicatesT, raw_filter, p_val_cutoff, target, 
                     model_formula_text, external_DEA_data, 
                     multifactorial)
#q()
    #################################################################
    ##                       CORRELATION ANALYSIS                   ##
    ##################################################################

    combinations_WGCNA <- NULL
    if(grepl("W", modules)) { # CASE W: WGCNA
      cat('Correlation analysis is performed with WGCNA\n')
      path <- file.path(output_files, "Results_WGCNA")
      dir.create(path)

      all_data_normalized <- exp_results[['all_data_normalized']]
      if(WGCNA_norm_method %in% names(all_data_normalized)) {
        WGCNA_input <- all_data_normalized[[WGCNA_norm_method]]
      } else if(WGCNA_norm_method== "none") {
        WGCNA_input <- raw_filter
      } else {
        warning(paste0("To run WGCNA, you must also run the method chosen for",
          " normalization in the --modules flag, or specify none"))
      }

      combinations_WGCNA <- perform_WGCNA_combinations(WGCNA_all, WGCNA_input, 
        index_treatmn_cols, index_control_cols, path, target_numeric_factors, 
        target_string_factors, WGCNA_memory, WGCNA_deepsplit, 
        WGCNA_detectcutHeight, WGCNA_mergecutHeight, WGCNA_min_genes_cluster, 
        WGCNA_blockwiseNetworkType, WGCNA_blockwiseTOMType, WGCNA_minCoreKME, 
        WGCNA_minCoreKMESize, WGCNA_minKMEtoStay)

      if(length(combinations_WGCNA[['WGCNA_all']]) == 1) {
        warning("Something went wrong with WGCNA on the full dataset")
        modules <- gsub("W", "", modules)
      }
    }

    if(grepl("X", modules)) { # CASE X: diffcoexp
      cat('Correlation analysis is performed with diffcoexp\n')
      path <- file.path(output_files, "Results_diffcoexp")
      dir.create(path)

      results_diffcoexp <- analysis_diffcoexp(data = raw_filter,
                                           path = path,
                                           target = target)
    }


    #################################################################
    ##                    BUILD MAIN RESULT TABLE                  ##
    #################################################################
    DE_all_genes <- unite_DEG_pack_results(exp_results, p_val_cutoff, 
                                           lfc, minpack_common)
    
    if(grepl("W", modules)) { # Check WGCNA was run and returned proper results
      DE_all_genes <- merge(by.x=0, by.y="ENSEMBL_ID", x= DE_all_genes, 
          y =combinations_WGCNA[['WGCNA_all']][['gene_cluster_info']])
      rownames(DE_all_genes) <- DE_all_genes$Row.names
      DE_all_genes$Row.names <- NULL
    }

    if(grepl("X", modules)) { #results_diffcoexp
      aux <- row.names(DE_all_genes)
      DE_all_genes$DCG <- aux %in% 
         results_diffcoexp$DCGs$Gene[results_diffcoexp$DCGs$q < 1]
      DE_all_genes$DCL <- aux %in% results_diffcoexp$DCLs$Gene.1 | 
         aux %in% results_diffcoexp$DCLs$Gene.2
    }
    
    # Add the filtered genes back
    DE_all_genes <- add_filtered_genes(DE_all_genes, raw)

    final_results <- list()
    final_results[['raw_filter']] <- raw_filter
    final_results[['sample_groups']] <- sample_groups
    final_results[['DE_all_genes']] <- DE_all_genes
    final_results[['index_control_cols']] <- index_control_cols
    final_results[['index_treatmn_cols']] <- index_treatmn_cols
    final_results[['design_vector']] <- design_vector
    final_results[['replicatesC']] <- replicatesC
    final_results[['replicatesT']] <- replicatesT
    final_results[['final_main_params']] <- final_main_params
    final_results[["var_filter"]] <- var_filter
    if(!is.null(combinations_WGCNA)){
      final_results <- c(final_results, combinations_WGCNA)
    }
    return(c(final_results, exp_results))
}



check_input_main_degenes_Hunter <- function(raw, 
                                            minlibraries, 
                                            reads, 
                                            external_DEA_data, 
                                            modules, 
                                            model_variables, 
                                            active_modules, 
                                            WGCNA_all, 
                                            minpack_common, 
                                            target, 
                                            string_factors, 
                                            numeric_factors,
                                            multifactorial){
    if (minlibraries < 1){
      stop(cat(paste0("Minimum library number to check minimum read counts",
        " cannot be less than 1.\nIf you want to avoid filtering, set",
        " --reads to 0.")))
    }

    if (reads == 0){
      cat(paste0("Minimum reads option is set to zero. ",
        "Raw count table will not be filtered."))
    }

    if (!is.null(external_DEA_data)) {
      warning("External DEA dataframe given.")
      if(is.null(target)){
        warning(paste0("External DEA dataframe given but no target",
          " - creating a blank one"))
        target <- data.frame(samples=c("case_dummy_1", "case_dummy_2", 
                                       "control_dummy_1", "control_dummy2"), 
                             treat=c("Treat","Treat", "Ctrl", "Ctrl"))
      }
      if(is.null(raw)) {
        warning(paste0("External DEA dataframe given but no count data",
          " - creating a blank one"))
        raw <- as.data.frame(matrix(1, 
                                    nrow=nrow(external_DEA_data), 
                                    ncol=nrow(target), 
                                  dimnames = list(row.names(external_DEA_data), 
                                                  target$sample)))
        reads <- 0
      }
    }

    if (model_variables != "" & grepl("N", modules) & nchar(modules) == 1) {
      stop(cat(paste0("You cannot run an experimental design that uses the",
        " model_variables option with NOISeq only.")))
    }
    if (model_variables != "" & grepl("N", modules) & nchar(modules) > 1) {
      warning(paste0("NOISeq will not be run as you have an experimental",
        " design that uses the model_variables option."))
      modules <- gsub("N", "", modules)
    }
    if(model_variables != "" & multifactorial != ""){
      stop("Cannot provide both a --multifactorial design and additional vars",
        "via --model_variables")
    }

    active_modules <- nchar(modules)
    if(grepl("W", modules)) {
      active_modules <- active_modules - 1
    }
    if(grepl("P", modules)) {
      if (WGCNA_all == FALSE) {
        message(paste0("WGCNA only for controls an treatments is not",
          " activated (--WGCNA_all option) and is needed for PCIT analysis.",
          " Disabling PCIT execution"))
        modules <- gsub("P", "", modules)
      } else {
        active_modules <- active_modules - 1
      }
    }
    if(grepl("X", modules)) {
      active_modules <- active_modules - 1
    }

    if(minpack_common > active_modules){
      minpack_common <- active_modules
      warning(paste0("The number of active modules is lower than the thresold",
        " for tag PREVALENT DEG. The thresold is set to the",
        " number of active modules."))
    }


    # If no -t check there is no -v value
    if( is.null(target) & model_variables != "") {
      stop(cat(paste0("You should not include a -v value if you do not",
        " have a target table file.")))
    }

    # If factors are specified but WGCNA not selected, throw a warning.
    if((string_factors != "" | numeric_factors != "") & 
       (!grepl("W", modules) | is.null(target))) {
      warning(paste0("If you wish to use factors for the correlation analysis",
        " you must also run WGCNA and include a target file. The -S",
        " and -N options will be ignored"))
    }
    return(list(modules=modules, 
                active_modules=active_modules, 
                minpack_common=minpack_common, 
                raw=raw, reads=reads, target=target))
}

#' @importFrom edgeR cpm
filter_count <- function(reads, 
                         minlibraries, 
                         raw, 
                         filter_type, 
                         index_control_cols, 
                         index_treatmn_cols){
    # Prepare filtered set
    if(reads != 0){
      if (filter_type == "separate") {
        # genes with cpm greater than --reads value for 
        # at least --minlibrariess samples for either case or control samples
        to_keep_control <- rowSums(edgeR::cpm(raw[index_control_cols]) > 
                                reads) >= minlibraries
        to_keep_treatment <- rowSums(edgeR::cpm(raw[index_treatmn_cols]) >
                                  reads) >= minlibraries
        # Keep if at least minlibraries in either of them
        keep_cpm <- to_keep_control | to_keep_treatment 
        raw <- raw[keep_cpm,] # Filter out count data frame
      } else if (filter_type == "global") {
        # genes with cpm greater than --reads value for
        # at least --minlibrariess samples
        keep_cpm <- rowSums(edgeR::cpm(raw) > reads) >= minlibraries 
        raw <- raw[keep_cpm,] # Filter out count data frame
      } else {
        warning("Unrecognized minimum read filter type. No filter will be used")
        raw <- raw
      }
    }
    return(raw)
}

#' @importFrom matrixStats rowVars
#' @importFrom stats quantile formula
#' @importFrom DESeq2 DESeqDataSetFromMatrix DESeq counts
filter_by_variance <- function(count_matrix, q_filter, target){

  dds <- DESeq2::DESeqDataSetFromMatrix(countData = count_matrix,
                                  colData = target,
                                  design = stats::formula("~ treat"))
  dds <- DESeq2::DESeq(dds)
  normalized_counts <-DESeq2::counts(dds, normalized=TRUE)
  variances <- matrixStats::rowVars(normalized_counts)
  threshold <- stats::quantile(variances, q_filter)
  fil_count_mtrx <- count_matrix[variances >= threshold,]
  return(list(fil_count_mtrx = fil_count_mtrx, 
    variance_dis = variances, 
     thr = q_filter))
}

prepare_model_text <- function(model_variables, 
                               multifactorial){
    # Prepare model text
    if(model_variables != "") {
      model_variables_unlist <- unlist(strsplit(model_variables, ","))
      model_formula_text <- paste("~", paste(model_variables_unlist, 
                                  "+", collapse=" "), "treat")
    } else if (multifactorial != "") {
      mf_text <- split_mf_text(multifactorial)
      mf_factor_A <- mf_text[[1]]
      mf_factor_B <- mf_text[[2]]
      model_formula_text <- paste0("~ ", mf_factor_B, " + ", 
                         mf_factor_A, " + ", mf_factor_B,":", mf_factor_A)
    } else {
      model_formula_text <- "~ treat"
    }
    cat("Model for gene expression analysis is:", model_formula_text, "\n")
    return(model_formula_text)
}

split_mf_text <- function(multifactorial) {
      factors_contrast <- strsplit(multifactorial,":")[[1]]
      factors <- strsplit(factors_contrast[1], ",")[[1]]
      mf_factorA <- factors[1]
      mf_factorB <- factors[2]
      contrast_varA_varB <- strsplit(factors_contrast[2],",")[[1]]
      mf_contrast <- contrast_varA_varB[1]
      mf_varA <- contrast_varA_varB[2]
      mf_varB <- contrast_varA_varB[3]
      return(list(mf_factorA=mf_factorA, mf_factorB=mf_factorB, 
        mf_contrast=mf_contrast, mf_varA=mf_varA, mf_varB=mf_varB))
}

prepare_target_for_multifactorial <- function(target, multifactorial) {
  mf_text <- split_mf_text(multifactorial)
  target[, mf_text[["mf_factorA"]]] <- stats::relevel(
                                             target[,mf_text[["mf_factorA"]]],
                                             mf_text[["mf_varA"]])
  target[, mf_text[["mf_factorB"]]] <- stats::relevel(
                                             target[,mf_text[["mf_factorB"]]],
                                             mf_text[["mf_varB"]])
  return(target)
}
